/**
  @file main.c
  @brief
  FT90x Ethernet to GPIO Bridge Example.

 */
/*
 * ============================================================================
 * History
 * =======
 * 12 May 2016 : Created
 *
 * (C) Copyright 2014-2016, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */


/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include <ft900.h>
#include <ft900_dlog.h>
#include <ft900_uart_simple.h>
#include <ft900_eth.h>
#include <ft900_startup_dfu.h>

/* UART support for printf output. */
#include "tinyprintf.h"

#include "net.h"

#include <lwip\apps\httpd.h>

#define BRIDGE_DEBUG
#ifdef BRIDGE_DEBUG
#define BRIDGE_DEBUG_PRINTF(...) do {tfp_printf(__VA_ARGS__);} while (0)
#else
#define BRIDGE_DEBUG_PRINTF(...)
#endif


/**
 @brief Link to datalogger area defined in crt0.S file.
 @details Must be passed to dlog library functions to initialise and use
        datalogger functions. We use the datalogger area for persistent
        configuration storage.
 */
extern __flash__ uint32_t __dlog_partition[];

/**
 @brief Pack structures to align on byte boundaries. Required for accurate layout
        of datalogger data structures.
 */
#define PACK __attribute__ ((__packed__))

/* CONSTANTS ***********************************************************************/

/**
 @brief Page number in datalogger memory for network configuration.
 */
#define NET_CONFIG_PAGE 0

/**
 @brief Page number in datalogger memory for GPIO configuration.
 */
#define GPIO_CONFIG_PAGE 1

/**
 @brief Magic number to determine if structure stored in datalogger memory
        for network configuration is valid.
 */
#define NET_CONFIG_KEY 0x46383912

/**
 @brief Magic number to determine if structure stored in datalogger memory
        for GPIO configuration is valid.
 */
#define GPIO_CONFIG_KEY 0x49382913


/* GLOBAL VARIABLES ****************************************************************/
uint8_t gpioout_val[4] = {0, 0, 0, 0};
uint8_t gpioin_val[4] = {0, 0, 0, 0};

/* LOCAL VARIABLES *****************************************************************/

/**
 @brief Area reserved to store configuration data which will be updated in
        datalogger storage area when required.
 */
static uint8_t	flashbuf[256] __attribute__((aligned(4)));

/**
 @brief Millisecond counter
 @details Count-up timer to provide the elapsed time for network operations.
 */
static uint32_t milliseconds = 0;

/**
 @brief lwIP timer counter
 @details Count-down timer to provide the a periodic 'tick' for lwIP timer
 	 operations.
 */
static uint16_t lwip_timer = 0;

static uint16_t stats_timer = 0;


/**
 @brief Structure to hold network settings.
 @details This is used to both keep current settings in use and
 	 to store persistently in Flash.
 */
struct net_config {
	uint32_t key;
	char hostname[64];
};

/**
 @brief Structure to hold GPIO settings.
 @details This is used to both keep current settings in use and
 	 to store persistently in Flash.
 */
struct gpio_config {
	uint32_t key;
	uint8_t gpioout_num[4];
	uint8_t gpioin_num[4];
	char outname[6];
	char outname1[6];
	char outname2[6];
	char outname3[6];
	char inname[6];
	char inname1[6];
	char inname2[6];
	char inname3[6];
};


/**
 @brief Default network configuration.
 */
static struct net_config default_net_config = {
		NET_CONFIG_KEY,
		// Hostname.
		{'G', 'P', 'I', 'O','B', 'r', 'i', 'd', 'g', 'e', 0, 0, 0,},
};

/**
 @brief Default GPIO configuration.
 */
static struct gpio_config default_gpio_config = {
		GPIO_CONFIG_KEY,
		{52, 53, 54, 55},
		{52, 53, 54, 55},
		{'M', 'O', 'T', 'O','R',0,},
		{'L', 'E', 'D', '1',0,0,},
		{'F', 'A', 'N', '_','A',0,},
		{'L', 'O', 'C', 'K','1',0,},
		{'D', 'O', 'O', 'R','1',0,},
		{'L', 'O', 'C', 'K','1',0,},
		{'L', 'I', 'G','H','T',0,},
		{'A', 'L', 'A', 'R','M',0,},
};

/**
 @brief Current network configuration.
 */
static struct net_config current_net_config;

/**
 @brief Current GPIO configuration.
 */
static struct gpio_config current_gpio_config;

/**
 @brief Watchdog keep alive flag.
 @details Set to zero to initiate a watchdog reset.
 */
int do_wdg_kick = 1;

/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

void flash_config_init(void);
void flash_config_write(void);
void flash_config_read(void);


/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

/** @brief Returns the current millisecond country
 *  @returns A count of milliseconds
 */
uint32_t millis(void)
{
	return milliseconds;
}

void ISR_timer(void)
{
	if (timer_is_interrupted(timer_select_a))
	{
		milliseconds++;

		if (lwip_timer)
		{
			lwip_timer--;
		}

		if (stats_timer)
		{
			stats_timer--;
		}
	}
}

// HTTPD Server Side Include Handler
#define FOREACH_SSI(MAKE_SSI) \
		MAKE_SSI(mac)   \
		MAKE_SSI(hn)   \
		MAKE_SSI(ip)  \
		MAKE_SSI(gw)   \
		MAKE_SSI(nm)  \
		MAKE_SSI(dhcp)  \
		MAKE_SSI(outnum)  \
		MAKE_SSI(outnum1)  \
		MAKE_SSI(outnum2)  \
		MAKE_SSI(outnum3)  \
		MAKE_SSI(outval)  \
		MAKE_SSI(outval1)  \
		MAKE_SSI(outval2)  \
		MAKE_SSI(outval3)  \
		MAKE_SSI(outname)  \
		MAKE_SSI(outname1)  \
		MAKE_SSI(outname2)  \
		MAKE_SSI(outname3)  \
		MAKE_SSI(innum)  \
		MAKE_SSI(innum1)  \
		MAKE_SSI(innum2)  \
		MAKE_SSI(innum3)  \
		MAKE_SSI(led)  \
		MAKE_SSI(led1)  \
		MAKE_SSI(led2)  \
		MAKE_SSI(led3)  \
		MAKE_SSI(inname)  \
		MAKE_SSI(inname1)  \
		MAKE_SSI(inname2)  \
		MAKE_SSI(inname3)  \
		MAKE_SSI(total)

#define GENERATE_SSI_ENUM(ENUM) ssiTag_##ENUM,
#define GENERATE_SSI_STRING(STRING) #STRING,

enum SSI_ENUM {
	FOREACH_SSI(GENERATE_SSI_ENUM)
};

static const char *ssiTags[] = {
		FOREACH_SSI(GENERATE_SSI_STRING)
};

u16_t app_httpd_ssi_cb(int iIndex, char *pcInsert, int iInsertLen)
{
	int len = 0;
	uint8_t *pval;
	ip_addr_t ia;
	char rstring[32];
	char *enstring = "";
	char *dstring;
	const char *nstring;
	struct eeprom_net_config getval;

	net_get_eeprom(&getval);

	if (getval.dhcp == 0)
	{
		enstring = "disabled";
	}

	if (iIndex < ssiTag_total)
	{
		nstring = ssiTags[iIndex];

		switch (iIndex)
		{
		case ssiTag_mac:
			pval = net_get_mac(rstring);
			sprintf(rstring, "%02x:%02x:%02x:%02x:%02x:%02x",
					pval[0], pval[1], pval[2], pval[3], pval[4], pval[5]);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" maxlength=\"18\" value=\"%s\" disabled>", nstring, rstring);
			break;
		case ssiTag_hn:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"64\">", nstring, current_net_config.hostname);
			break;
		case ssiTag_ip:
			ia = getval.ip;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_gw:
			ia = getval.gw;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_nm:
			ia = getval.mask;
			dstring = ipaddr_ntoa(&ia);
			len = sprintf(pcInsert, " <input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"16\" %s>", nstring, dstring, enstring);
			break;
		case ssiTag_dhcp:
			if (getval.dhcp)
			{
				dstring = "checked";
			}
			len = sprintf(pcInsert, " <input type=\"checkbox\" name=\"%s\" %s>", nstring, dstring);
			break;
		case ssiTag_outnum:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioout_num[0]);
			break;
		case ssiTag_outval:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"1\">", nstring, gpioout_val[0]);
			break;
		case ssiTag_outnum1:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioout_num[1]);
			break;
		case ssiTag_outval1:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"1\">", nstring, gpioout_val[1]);
			break;
		case ssiTag_outnum2:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioout_num[2]);
			break;
		case ssiTag_outval2:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"1\">", nstring, gpioout_val[2]);
			break;
		case ssiTag_outnum3:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioout_num[3]);
			break;
		case ssiTag_outval3:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"1\">", nstring, gpioout_val[3]);
			break;
		case ssiTag_outname:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.outname);
			break;
		case ssiTag_outname1:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.outname1);
			break;
		case ssiTag_outname2:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.outname2);
			break;
		case ssiTag_outname3:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.outname3);
			break;
		case ssiTag_innum:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioin_num[0]);
			break;
		case ssiTag_led:
			gpioin_val[0] = gpio_read(current_gpio_config.gpioin_num[0]);
				if (gpioin_val[0] == 1)
				{
					len = sprintf(pcInsert, "<img src=\"img/on.png\" border=\"0\">");
				}
				else
				{
					len = sprintf(pcInsert, "<img src=\"img/off.png\" border=\"0\">");
				}
		break;
		case ssiTag_innum1:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioin_num[1]);
			break;
		case ssiTag_led1:
			gpioin_val[1] = gpio_read(current_gpio_config.gpioin_num[1]);
				if (gpioin_val[1] == 1)
				{
					len = sprintf(pcInsert, "<img src=\"img/on.png\" border=\"0\">");
				}
				else
				{
					len = sprintf(pcInsert, "<img src=\"img/off.png\" border=\"0\">");
				}
		break;
		case ssiTag_innum2:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioin_num[2]);
			break;
		case ssiTag_led2:
			gpioin_val[2] = gpio_read(current_gpio_config.gpioin_num[2]);
				if (gpioin_val[2] == 1)
				{
					len = sprintf(pcInsert, "<img src=\"img/on.png\" border=\"0\">");
				}
				else
				{
					len = sprintf(pcInsert, "<img src=\"img/off.png\" border=\"0\">");
				}
		break;
		case ssiTag_innum3:
			len = sprintf(pcInsert, "<input type=\"number\" name=\"%s\" value=\"%d\" size=\"2\" min=\"0\" max=\"66\">", nstring, current_gpio_config.gpioin_num[3]);
			break;
		case ssiTag_led3:
			gpioin_val[3] = gpio_read(current_gpio_config.gpioin_num[3]);
				if (gpioin_val[3] == 1)
				{
					len = sprintf(pcInsert, "<img src=\"img/on.png\" border=\"0\">");
				}
				else
				{
					len = sprintf(pcInsert, "<img src=\"img/off.png\" border=\"0\">");
				}
		break;
		case ssiTag_inname:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.inname);
			break;
		case ssiTag_inname1:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.inname1);
			break;
		case ssiTag_inname2:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.inname2);
			break;
		case ssiTag_inname3:
			len = sprintf(pcInsert, "<input type=\"text\" name=\"%s\" value=\"%s\" maxlength=\"5\" size=\"5\">", nstring, current_gpio_config.inname3);
			break;
		}
	}

	return len;
}

// HTTPD CGI Handlers

const char *app_httpd_cgi_net(int iIndex, int iNumParams, char *pcParam[],
		char *pcValue[])
{
	int i;
	int mac[6];
	int val;
	struct eeprom_net_config setval;

	// Checkboxes not always sent if unchecked.
	net_get_eeprom(&setval);
	setval.dhcp = 0;

	for(i = 0; i < iNumParams; i++)
	{
		if(!strcmp(pcParam[i], ssiTags[ssiTag_mac]))
		{
			if (sscanf(pcValue[i], "%x:%x:%x:%x:%x:%x",
			    &mac[0], &mac[1], &mac[2],
			    &mac[3], &mac[4], &mac[5]) == 6)
			{
			    for (val = 0; val < 6; val++)
			    {
			        //current_net_config.macaddr[val] = (uint8_t) mac[val];
			    }
			}
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_hn]))
		{
			strcpy(current_net_config.hostname, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_dhcp]))
		{
			setval.dhcp = 1;
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_ip]))
		{
			ipaddr_aton(pcValue[i], &setval.ip);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_gw]))
		{
			ipaddr_aton(pcValue[i], &setval.gw);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_nm]))
		{
			ipaddr_aton(pcValue[i], &setval.mask);
		}
	}

	BRIDGE_DEBUG_PRINTF("Net Config:\n");
	//BRIDGE_DEBUG_PRINTF("  MAC = %02x:%02x:%02x:%02x:%02x:%02x\n",
		//	current_net_config.macaddr[0], current_net_config.macaddr[1],
		//	current_net_config.macaddr[2], current_net_config.macaddr[3],
		//	current_net_config.macaddr[4], current_net_config.macaddr[5]);
	BRIDGE_DEBUG_PRINTF("  Hostname = \"%s\"\n", current_net_config.hostname);
	BRIDGE_DEBUG_PRINTF("  DHCP = %c\n", setval.dhcp?'Y':'N');
	BRIDGE_DEBUG_PRINTF("  ipaddr = %s\n", ipaddr_ntoa(&setval.ip));
	BRIDGE_DEBUG_PRINTF("  gwaddr = %s\n", ipaddr_ntoa(&setval.gw));
	BRIDGE_DEBUG_PRINTF("  netmask = %s\n", ipaddr_ntoa(&setval.mask));

	net_update_eeprom(&setval);
	flash_config_write();

	// Force a reboot through the watchdog timer.
	do_wdg_kick = 0;

	return "/restart.shtm";
}

const char *app_httpd_cgi_gpioout(int iIndex, int iNumParams, char *pcParam[],
		char *pcValue[])
{
	int i;

	for(i = 0; i < iNumParams; i++)
	{
		if(!strcmp(pcParam[i], ssiTags[ssiTag_outnum]))
		{
		 current_gpio_config.gpioout_num[0] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioout_num[0], pad_dir_output);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outval]))
		{
		 gpioout_val[0] = atoi(pcValue[i]);
	     gpio_write(current_gpio_config.gpioout_num[0], gpioout_val[0]);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_outnum1]))
		{
		 current_gpio_config.gpioout_num[1] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioout_num[1], pad_dir_output);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outval1]))
		{
		 gpioout_val[1] = atoi(pcValue[i]);
	     gpio_write(current_gpio_config.gpioout_num[1], gpioout_val[1]);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_outnum2]))
		{
		 current_gpio_config.gpioout_num[2] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioout_num[2], pad_dir_output);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outval2]))
		{
		 gpioout_val[2] = atoi(pcValue[i]);
	     gpio_write(current_gpio_config.gpioout_num[2], gpioout_val[2]);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_outnum3]))
		{
		 current_gpio_config.gpioout_num[3] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioout_num[3], pad_dir_output);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outval3]))
		{
		 gpioout_val[3] = atoi(pcValue[i]);
	     gpio_write(current_gpio_config.gpioout_num[3], gpioout_val[3]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outname]))
		{
		 strcpy(current_gpio_config.outname, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outname1]))
		{
		 strcpy(current_gpio_config.outname1, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outname2]))
		{
		 strcpy(current_gpio_config.outname2, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_outname3]))
		{
		 strcpy(current_gpio_config.outname3, pcValue[i]);
		}
	}

	flash_config_write();

	return "/index.shtm";
}

const char *app_httpd_cgi_gpioin(int iIndex, int iNumParams, char *pcParam[],
		char *pcValue[])
{
	int i;

	for(i = 0; i < iNumParams; i++)
	{
		if(!strcmp(pcParam[i], ssiTags[ssiTag_innum]))
		{
		 current_gpio_config.gpioin_num[0] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioin_num[0], pad_dir_input);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_innum1]))
		{
		 current_gpio_config.gpioin_num[1] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioin_num[1], pad_dir_input);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_innum2]))
		{
		 current_gpio_config.gpioin_num[2] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioin_num[2], pad_dir_input);
		}
		if(!strcmp(pcParam[i], ssiTags[ssiTag_innum3]))
		{
		 current_gpio_config.gpioin_num[3] = atoi(pcValue[i]);
		 gpio_dir(current_gpio_config.gpioin_num[3], pad_dir_input);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_inname]))
		{
		 strcpy(current_gpio_config.inname, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_inname1]))
		{
		 strcpy(current_gpio_config.inname1, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_inname2]))
		{
		 strcpy(current_gpio_config.inname2, pcValue[i]);
		}
		else if(!strcmp(pcParam[i], ssiTags[ssiTag_inname3]))
		{
		 strcpy(current_gpio_config.inname3, pcValue[i]);
		}
	}

	flash_config_write();

	return "/index.shtm";
}

const tCGI cgiHandler[] = {
		{"/net", app_httpd_cgi_net},
		{"/gpioout", app_httpd_cgi_gpioout},
		{"/gpioin", app_httpd_cgi_gpioin},
};

void flash_config_init(void)
{
	int	pgsz;
	int	i;

	if (dlog_init(__dlog_partition, &pgsz, &i) < 0)
	{
		// Project settings incorrect. Require dlog support with modified
		// linker script and crt0.S file.
		// See AN_398 for examples.
		return;
	}
}

void flash_config_write(void)
{
	dlog_erase();
	memset(flashbuf, 0xff, sizeof(flashbuf));
	memcpy(flashbuf, &current_net_config, sizeof(struct net_config));
	dlog_prog(NET_CONFIG_PAGE, (uint32_t *)flashbuf);
	memset(flashbuf, 0xff, sizeof(flashbuf));
	memcpy(flashbuf, &current_gpio_config, sizeof(struct gpio_config));
	dlog_prog(GPIO_CONFIG_PAGE, (uint32_t *)flashbuf);
}

void flash_config_read(void)
{
	dlog_read(NET_CONFIG_PAGE, (uint32_t *)flashbuf);
	memcpy(&current_net_config, flashbuf, sizeof(struct net_config));
	dlog_read(GPIO_CONFIG_PAGE, (uint32_t *)flashbuf);
	memcpy(&current_gpio_config, flashbuf, sizeof(struct gpio_config));
}

void gpio_bridge(void)
{
	ip_addr_t ip_addr = {0};
	uint8_t flash_init = 0;

	flash_config_init();
	flash_config_read();

	// Implement defaults if required.
	if (current_net_config.key != NET_CONFIG_KEY)
	{
		BRIDGE_DEBUG_PRINTF("Using default network configuration\n");
		// Setup using default network configuration.
		memcpy(&current_net_config, &default_net_config, sizeof(struct net_config));
		flash_init = 1;
	}
	if (current_gpio_config.key != GPIO_CONFIG_KEY)
	{
		BRIDGE_DEBUG_PRINTF("Using default GPIO configuration\n");
		// Setup using default GPIO configuration.
		memcpy(&current_gpio_config, &default_gpio_config, sizeof(struct gpio_config));
		flash_init = 1;
	}
	if (flash_init != 0)
	{
		flash_config_write();
	}

	wdt_kick();

	// Open the lwIP library and ask for a DHCP address.
	// Open the lwIP library and existing EEPROM settings.
	if (net_init(ip_addr, ip_addr, ip_addr, 0, current_net_config.hostname) == ERR_OK)
	{
		int started = 0;

		while (1)
		{
			if (net_is_link_up())
			{
				net_tick();

				if (!started)
				{
					started = net_is_up();
					if (started)
					{
						httpd_init();
						http_set_ssi_handler(app_httpd_ssi_cb, ssiTags, ssiTag_total);
						http_set_cgi_handlers(cgiHandler, sizeof(cgiHandler)/sizeof(tCGI));
					}
				}
			}

			if (do_wdg_kick)
			{
				wdt_kick();
			}

			if (stats_timer == 0)
			{
				//stats_display();
				//netif_stats_display_ft900();
				stats_timer = 10000;
			}

			if (lwip_timer == 0)
			{
				// Check for ethernet disconnection.
				if (net_is_link_up())
				{
					if (!ethernet_is_link_up())
					{
						net_set_link_down();
						BRIDGE_DEBUG_PRINTF("Ethernet disconnected.\n");
					}
				}
				else
				{
					if (ethernet_is_link_up())
					{
						net_set_link_up();
						BRIDGE_DEBUG_PRINTF("Ethernet connected.\n");
					}
				}

				// Check lwIP timeouts every 250 ms.
				sys_check_timeouts();
				lwip_timer = 250;
			}
		}
	}
	// Disconnected from host.
}

/* FUNCTIONS ***********************************************************************/

int main(void)
{
	sys_reset_all();
    sys_disable(sys_device_i2c_master);

	STARTUP_DFU();

	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
	/* Make GPIO48 function as UART0_TXD and GPIO49 function as UART0_RXD... */
	gpio_function(48, pad_uart0_txd); /* UART0 TXD */
	gpio_function(49, pad_uart0_rxd); /* UART0 RXD */

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);

#ifdef BRIDGE_DEBUG
	// Open the UART using the coding required.
	uart_open(UART0, 1, UART_DIVIDER_19200_BAUD, uart_data_bits_8, uart_parity_none, uart_stop_bits_1);

	/* Print out a welcome message... */
	BRIDGE_DEBUG_PRINTF("\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	BRIDGE_DEBUG_PRINTF("(C) Copyright 2016, Future Technology Devices International Ltd. \r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
	BRIDGE_DEBUG_PRINTF("Welcome to the Ethernet GPIO Bridge ... \r\n");
	BRIDGE_DEBUG_PRINTF("\r\n");
	BRIDGE_DEBUG_PRINTF("Connect to FT90x via Ethernet and control GPIO.\r\n");
	BRIDGE_DEBUG_PRINTF("--------------------------------------------------------------------- \r\n");
#endif // BRIDGE_DEBUG

	sys_enable(sys_device_timer_wdt);

/* Timer A = 1ms */
	timer_prescaler(1000);
	timer_init(timer_select_a, 100, timer_direction_down, timer_prescaler_select_on, timer_mode_continuous);
	timer_enable_interrupt(timer_select_a);
	timer_start(timer_select_a);

	interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, ISR_timer);

	// Start watchdog timer to restart the device after approx 1.34 seconds.
	wdt_init(wdt_counter_1G_clocks);

    i2cm_init(I2CM_NORMAL_SPEED, 10000);

    sys_enable(sys_device_ethernet);

    /* Set up I2C */
    sys_enable(sys_device_i2c_master);

    /* Setup I2C channel 0 pins */
    /* Use sys_i2c_swop(0) to activate. */
    gpio_function(44, pad_i2c0_scl); /* I2C0_SCL */
    gpio_function(45, pad_i2c0_sda); /* I2C0_SDA */

    /* Setup I2C channel 1 pins for EEPROM */
    /* Use sys_i2c_swop(1) to activate. */
    gpio_function(46, pad_i2c1_scl); /* I2C1_SCL */
    gpio_function(47, pad_i2c1_sda); /* I2C1_SDA */

    interrupt_enable_globally();

	gpio_bridge();

	interrupt_disable_globally();

	// Wait forever...
	for (;;);

	return 0;
}
